@@ -12,12 +12,12 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with g15lcd; if not, write to the Free Software
+    along with libg15; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    (c) 2006 - 2007 The G15tools Project - g15tools.sf.net
+    (c) 2006-2007 The G15tools Project - g15tools.sf.net
     
-    $Revision: 260 $ -  $Date: 2007-10-18 12:17:48 +0930 (Thu, 18 Oct 2007) $ $Author: mlampard $
+    $Revision: 282 $ -  $Date: 2008-01-20 06:32:00 +1030 (Sun, 20 Jan 2008) $ $Author: mlampard $
 */
 
 #include <pthread.h>
@@ -27,6 +27,7 @@
 #include <usb.h>
 #include <string.h>
 #include <errno.h>
+#include "config.h"
 
 static usb_dev_handle *keyboard_device = 0;
 static int libg15_debugging_enabled = 0;
@@ -38,14 +39,14 @@ static int g15_keys_endpoint = 0;
 static int g15_lcd_endpoint = 0;
 static pthread_mutex_t libusb_mutex;
 
-
 /* to add a new device, simply create a new DEVICE() in this list */
 /* Fields are: "Name",VendorID,ProductID,Capabilities */
 const libg15_devices_t g15_devices[] = {
     DEVICE("Logitech G15",0x46d,0xc222,G15_LCD|G15_KEYS),
     DEVICE("Logitech G11",0x46d,0xc225,G15_KEYS),
     DEVICE("Logitech Z-10",0x46d,0x0a07,G15_LCD|G15_KEYS|G15_DEVICE_IS_SHARED),
-    DEVICE("Logitech G15 v2",0x46d,0xc227,G15_LCD|G15_KEYS),
+    DEVICE("Logitech G15 v2",0x46d,0xc227,G15_LCD|G15_KEYS|G15_DEVICE_5BYTE_RETURN),
+    DEVICE("Logitech Gamepanel",0x46d,0xc251,G15_LCD|G15_KEYS|G15_DEVICE_IS_SHARED),
     DEVICE(NULL,0,0,0)
 };
 
@@ -196,7 +197,7 @@ static usb_dev_handle * findAndOpenDevice(libg15_devices_t handled_device, int d
                                     }
                                     else
                                     {
-                                        g15_log(stderr,G15_LOG_INFO,"Sorry, I could not detached the driver, giving up\n");
+                                        g15_log(stderr,G15_LOG_INFO,"Sorry, I could not detach the driver, giving up\n");
                                         return 0;
                                     }
 
@@ -301,7 +302,13 @@ int initLibG15()
     retval = initLibUsb();
     if (retval)
         return retval;
-  
+
+    g15_log(stderr,G15_LOG_INFO,"%s\n",PACKAGE_STRING);
+    
+#ifdef SUN_LIBUSB
+    g15_log(stderr,G15_LOG_INFO,"Using Sun libusb.\n");
+#endif
+
     g15NumberOfConnectedDevices();
   
     keyboard_device = findAndOpenG15();
@@ -317,8 +324,10 @@ int exitLibG15()
 {
     int retval = G15_NO_ERROR;
     if (keyboard_device){
+#ifndef SUN_LIBUSB
         retval = usb_release_interface (keyboard_device, 0);
         usleep(50*1000);
+#endif
 #if 0
         retval = usb_reset(keyboard_device);
         usleep(50*1000);
@@ -334,36 +343,84 @@ int exitLibG15()
 
 static void dumpPixmapIntoLCDFormat(unsigned char *lcd_buffer, unsigned char const *data)
 {
-    unsigned int offset_from_start = G15_LCD_OFFSET;
+/*
+
+  For a set of bytes (A, B, C, etc.) the bits representing pixels will appear on the LCD like this:
+	
+	A0 B0 C0
+	A1 B1 C1
+	A2 B2 C2
+	A3 B3 C3 ... and across for G15_LCD_WIDTH bytes
+	A4 B4 C4
+	A5 B5 C5
+	A6 B6 C6
+	A7 B7 C7
+	
+	A0
+	A1  <- second 8-pixel-high row starts straight after the last byte on
+	A2     the previous row
+	A3
+	A4
+	A5
+	A6
+	A7
+	A8
+
+	A0
+	...
+	A0
+	...
+	A0
+	...
+	A0
+	A1 <- only the first three bits are shown on the bottom row (the last three
+	A2    pixels of the 43-pixel high display.)
+	
+
+*/
+
+    unsigned int output_offset = G15_LCD_OFFSET;
+    unsigned int base_offset = 0;
     unsigned int curr_row = 0;
     unsigned int curr_col = 0;
-  
-    for (curr_row=0;curr_row<G15_LCD_HEIGHT;++curr_row)
-    {
-        for (curr_col=0;curr_col<G15_LCD_WIDTH;++curr_col)
-        {
-            unsigned int pixel_offset = curr_row*G15_LCD_WIDTH + curr_col;
-            unsigned int byte_offset = pixel_offset / 8;
-            unsigned int bit_offset = pixel_offset % 8;
-            unsigned int val = data[byte_offset] & 1<<(7-bit_offset);
-      
-            unsigned int row = curr_row / 8;
-            unsigned int offset = G15_LCD_WIDTH*row + curr_col;
-            unsigned int bit = curr_row % 8;
-    
-/*
-            if (val)
-            printf("Setting pixel at row %d col %d to %d offset %d bit %d\n",curr_row,curr_col, val, offset, bit);
+
+    /* Five 8-pixel rows + a little 3-pixel row.  This formula will calculate
+       the minimum number of bytes required to hold a complete column.  (It
+       basically divides by eight and rounds up the result to the nearest byte,
+       but at compile time.
       */
-            if (val)
-                lcd_buffer[offset_from_start + offset] = lcd_buffer[offset_from_start + offset] | 1 << bit;
-            else
-                lcd_buffer[offset_from_start + offset] = lcd_buffer[offset_from_start + offset]  &  ~(1 << bit);
+
+#define G15_LCD_HEIGHT_IN_BYTES  ((G15_LCD_HEIGHT + ((8 - (G15_LCD_HEIGHT % 8)) % 8)) / 8)
+
+    for (curr_row = 0; curr_row < G15_LCD_HEIGHT_IN_BYTES; ++curr_row)
+    {
+        for (curr_col = 0; curr_col < G15_LCD_WIDTH; ++curr_col)
+        {
+            unsigned int bit = curr_col % 8;
+		/* Copy a 1x8 column of pixels across from the source image to the LCD buffer. */
+		
+            lcd_buffer[output_offset] =
+			(((data[base_offset                        ] << bit) & 0x80) >> 7) |
+			(((data[base_offset +  G15_LCD_WIDTH/8     ] << bit) & 0x80) >> 6) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 2)] << bit) & 0x80) >> 5) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 3)] << bit) & 0x80) >> 4) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 4)] << bit) & 0x80) >> 3) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 5)] << bit) & 0x80) >> 2) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 6)] << bit) & 0x80) >> 1) |
+			(((data[base_offset + (G15_LCD_WIDTH/8 * 7)] << bit) & 0x80) >> 0);
+            ++output_offset;
+            if (bit == 7)
+              base_offset++;
         }
+	/* Jump down seven pixel-rows in the source image, since we've just
+	   done a row of eight pixels in one pass (and we counted one pixel-row
+  	   while we were going, so now we skip the next seven pixel-rows.) */
+	base_offset += G15_LCD_WIDTH - (G15_LCD_WIDTH / 8);
     }
 }
 
 int handle_usb_errors(const char *prefix, int ret) {
+
     switch (ret){
         case -ETIMEDOUT:
             return G15_ERROR_READING_USB_DEVICE;  /* backward-compatibility */
@@ -378,7 +435,7 @@ int handle_usb_errors(const char *prefix, int ret) {
             case -EAGAIN: /* try again */
             case -EFBIG: /* too many frames to handle */
             case -EMSGSIZE: /* msgsize is invalid */
-                 g15_log(stderr,G15_LOG_INFO,"usb error: %s (%i)\n",prefix,ret);     
+                 g15_log(stderr,G15_LOG_INFO,"usb error: %s %s (%i)\n",prefix,usb_strerror(),ret);     
                  break;
             case -EPIPE: /* endpoint is stalled */
                  g15_log(stderr,G15_LOG_INFO,"usb error: %s EPIPE! clearing...\n",prefix);     
@@ -387,7 +444,7 @@ int handle_usb_errors(const char *prefix, int ret) {
                  pthread_mutex_unlock(&libusb_mutex);
                  break;
             default: /* timed out */
-                 g15_log(stderr,G15_LOG_INFO,"Unknown usb error: %s !! (err is %i)\n",prefix,ret);     
+                 g15_log(stderr,G15_LOG_INFO,"Unknown usb error: %s !! (err is %i (%s))\n",prefix,ret,usb_strerror());
     }
     return ret;
 }
@@ -397,7 +454,10 @@ int writePixmapToLCD(unsigned char const *data)
     int ret = 0;
     int transfercount=0;
     unsigned char lcd_buffer[G15_BUFFER_LEN];
-    memset(lcd_buffer,0,G15_BUFFER_LEN);
+    /* The pixmap conversion function will overwrite everything after G15_LCD_OFFSET, so we only need to blank
+       the buffer up to this point.  (Even though the keyboard only cares about bytes 0-23.) */
+    memset(lcd_buffer, 0, G15_LCD_OFFSET);  /* G15_BUFFER_LEN); */
+
     dumpPixmapIntoLCDFormat(lcd_buffer, data);
 
     if(!(g15_devices[found_devicetype].caps & G15_LCD))
@@ -411,7 +471,9 @@ int writePixmapToLCD(unsigned char const *data)
     I'm not sure how successful this will be in combatting ENOSPC, but we'll give it try in the real-world. */
 
     if(enospc_slowdown != 0){
+#ifndef LIBUSB_BLOCKS
         pthread_mutex_lock(&libusb_mutex);  
+#endif
         for(transfercount = 0;transfercount<=31;transfercount++){
             ret = usb_interrupt_write(keyboard_device, g15_lcd_endpoint, (char*)lcd_buffer+(32*transfercount), 32, 1000);
             if (ret != 32)
@@ -421,12 +483,18 @@ int writePixmapToLCD(unsigned char const *data)
             }
             usleep(100);
         }
+#ifndef LIBUSB_BLOCKS
         pthread_mutex_unlock(&libusb_mutex);
+#endif        
     }else{
         /* transfer entire buffer in one hit */
+#ifdef LIBUSB_BLOCKS
+        ret = usb_interrupt_write(keyboard_device, g15_lcd_endpoint, (char*)lcd_buffer, G15_BUFFER_LEN, 1000);
+#else
         pthread_mutex_lock(&libusb_mutex);
         ret = usb_interrupt_write(keyboard_device, g15_lcd_endpoint, (char*)lcd_buffer, G15_BUFFER_LEN, 1000);
         pthread_mutex_unlock(&libusb_mutex);
+#endif
         if (ret != G15_BUFFER_LEN)
         {
             handle_usb_errors ("LCDPixmap Write",ret);
@@ -543,7 +611,7 @@ static unsigned char g15KeyToLogitechKeyCode(int key)
     }
 }
 
-static void processKeyEvent(unsigned int *pressed_keys, unsigned char *buffer)
+static void processKeyEvent9Byte(unsigned int *pressed_keys, unsigned char *buffer)
 {
     int i;
   
@@ -634,22 +702,96 @@ static void processKeyEvent(unsigned int *pressed_keys, unsigned char *buffer)
     }
 }
 
+static void processKeyEvent5Byte(unsigned int *pressed_keys, unsigned char *buffer)
+{
+    int i;
+  
+    *pressed_keys = 0;
+  
+    g15_log(stderr,G15_LOG_WARN,"Keyboard: %x, %x, %x, %x, %x\n",buffer[0],buffer[1],buffer[2],buffer[3],buffer[4]);
+  
+    if (buffer[0] == 0x02)
+    {
+        if (buffer[1]&0x01)
+            *pressed_keys |= G15_KEY_G1;
+    
+        if (buffer[1]&0x02)
+            *pressed_keys |= G15_KEY_G2;
+
+        if (buffer[1]&0x04)
+            *pressed_keys |= G15_KEY_G3;
+
+        if (buffer[1]&0x08)
+            *pressed_keys |= G15_KEY_G4;
+
+        if (buffer[1]&0x10)
+            *pressed_keys |= G15_KEY_G5;
+
+        if (buffer[1]&0x20)
+            *pressed_keys |= G15_KEY_G6;
+            
+        if (buffer[1]&0x40)
+            *pressed_keys |= G15_KEY_M1;
+            
+        if (buffer[1]&0x80)
+            *pressed_keys |= G15_KEY_M2;
+            
+        if (buffer[2]&0x20)
+            *pressed_keys |= G15_KEY_M3;
+            
+        if (buffer[2]&0x40)
+            *pressed_keys |= G15_KEY_MR;
+
+        if (buffer[2]&0x80)
+            *pressed_keys |= G15_KEY_L1;
+            
+        if (buffer[2]&0x2)
+            *pressed_keys |= G15_KEY_L2;
+            
+        if (buffer[2]&0x4)
+            *pressed_keys |= G15_KEY_L3;
+
+        if (buffer[2]&0x8)
+            *pressed_keys |= G15_KEY_L4;
+            
+        if (buffer[2]&0x10)
+            *pressed_keys |= G15_KEY_L5;
+            
+        if (buffer[2]&0x1)
+            *pressed_keys |= G15_KEY_LIGHT;
+    }
+}
+
+
 int getPressedKeys(unsigned int *pressed_keys, unsigned int timeout)
 {
     unsigned char buffer[9];
     int ret = 0;
+#ifdef LIBUSB_BLOCKS
+    ret = usb_interrupt_read(keyboard_device, g15_keys_endpoint, (char*)buffer, 9, timeout);
+#else
     pthread_mutex_lock(&libusb_mutex);
     ret = usb_interrupt_read(keyboard_device, g15_keys_endpoint, (char*)buffer, 9, timeout);
     pthread_mutex_unlock(&libusb_mutex);
-   
-    if (ret == 9)
-    {
-        if (buffer[0] == 1)
-            return G15_ERROR_TRY_AGAIN;
+#endif
+    if(ret>0) {
+      if(buffer[0] == 1)
+        return G15_ERROR_TRY_AGAIN;    
+    }
+
+    if(g15DeviceCapabilities() & G15_DEVICE_5BYTE_RETURN) {
+      processKeyEvent5Byte(pressed_keys, buffer);
+      return G15_NO_ERROR;  
+    }
     
-        processKeyEvent(pressed_keys, buffer);
-    
-        return G15_NO_ERROR;
+    switch(ret) {
+      case 5:
+          processKeyEvent5Byte(pressed_keys, buffer);
+          return G15_NO_ERROR;
+      case 9:
+          processKeyEvent9Byte(pressed_keys, buffer);
+          return G15_NO_ERROR;
+      default:
+          return handle_usb_errors("Keyboard Read", ret); /* allow the app to deal with errors */
     }
-    return handle_usb_errors("Keyboard Read", ret); /* allow the app to deal with errors */
 }
